@@ -6,6 +6,10 @@
     See the file COPYING.LIB.
 */
 
+/* For pthread_rwlock_t */
+#define _GNU_SOURCE
+
+#include <config.h>
 #include "fuse.h"
 #include <pthread.h>
 
@@ -25,10 +29,15 @@ struct fuse {
     unsigned int hidectr;
     pthread_mutex_t lock;
     pthread_mutex_t worker_lock;
+    pthread_rwlock_t tree_lock;
     int numworker;
     int numavail;
     volatile int exited;
     int got_init;
+    void *user_data;
+    int major;
+    int minor;
+    uid_t owner;
 };
 
 struct fuse *fuse_new_common(int fd, const char *opts,
