@@ -44,7 +44,9 @@ static void usage(const char *progname)
             "    hard_remove            immediate removal (don't hide files)\n"
             "    debug                  enable debug output\n"
             "    fsname=NAME            set filesystem name in mtab\n"
-            "    use_ino                let filesystem set inode numbers\n");
+            "    use_ino                let filesystem set inode numbers\n"
+            "    readdir_ino            try to fill in d_ino in readdir\n"
+            );
 }
 
 static void invalid_option(const char *argv[], int argctr)
@@ -138,8 +140,12 @@ static int add_options(char **lib_optp, char **kernel_optp, const char *opts)
 
     while((opt = strsep(&s, ",")) != NULL) {
         int res;
-        if (fuse_is_lib_option(opt))
+        if (fuse_is_lib_option(opt)) {
             res = add_option_to(opt, lib_optp);
+            /* Compatibility hack */
+            if (strcmp(opt, "allow_root") == 0 && res != -1)
+                res = add_option_to("allow_other", kernel_optp);
+        }
         else
             res = add_option_to(opt, kernel_optp);
         if (res == -1) {
